// This file implements the CAI dynamic programming algorithm.
// 
// Codon Optimizer - a reference implementation of the CAI optimization 
// and motif engineering algorithm by Anne Condon and Chris Thachuk. 
//
// Copyright (C) 2012 Chris Thachuk
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#ifndef _CAI_OPTIMIZER_H_
#define _CAI_OPTIMIZER_H_

#include <cmath>
#include <iostream>
#include <map>
#include <string>
#include <vector>

#include "ac_matcher.h"
#include "codon.h"
#include "codon_table.h"
#include "const.h"
#include "counter.h"
#include "dna_sequence.h"

namespace oligodesign {

namespace {

int BuildACMatcher(ACMatcher<int, 4> &matcher,
                   const std::vector<DNASequence> &motifs) {
  int motifs_max_len = 0;

  std::vector<DNASequence>::const_iterator itr;
  for (itr = motifs.begin(); itr != motifs.end(); ++itr) {
    const DNASequence &dna = *itr;
    int length = dna.Length();
    motifs_max_len = std::max(motifs_max_len, length);
    matcher.AddString(&(dna[0]), length);
  }
  matcher.MakeTree();
  return motifs_max_len;
}

} // end anonymous namespace

int CountMotifOccurences(const DNASequence &seq, const std::vector<DNASequence> &motifs) {
  const int *dna_seq = &seq[0];
  int length = seq.Length();

  ACMatcher<int, 4> m_matcher;
  BuildACMatcher(m_matcher, motifs);
  int num_matches = m_matcher.Search(dna_seq, length);
  
  return num_matches;
}

double CalcCAI(const DNASequence &dna, const CodonTable &ctable) {
  int codonFreq[4][4][4];

  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < 4; j++) {
      for (int k = 0; k < 4; k++) {
        codonFreq[i][j][k] = 0;
      }
    }
  }

  int tot_aa = dna.Length() / 3;
  for (int i = 0; i < tot_aa; i++) {
    codonFreq[dna[i*3]][dna[3*i+1]][dna[3*i+2]]++;
  }

  double cai_numerator = 0.0;
  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < 4; j++) {
      for (int k = 0; k < 4; k++) {
        cai_numerator += codonFreq[i][j][k] * std::log(ctable.GetFit(i, j, k));
      }
    }
  }

  return std::exp(cai_numerator / (double)tot_aa);
}

double OptimizeCAI(DNASequence &dna, const CodonTable &ctable) {
  int length = dna.Length();
  for (int i = 0; i < length; i += 3) {
    ctable.SwapBest(&dna[i]);
  }
  return 1.0;
}

double OptimizeCAI(DNASequence &dna, const CodonTable &ctable,
                   std::vector<DNASequence> &forbidden,
                   std::vector<DNASequence> &desired) {
  // Modifed by GENETYX 2021.08.02
  if (forbidden.empty() && desired.empty()) {
    return OptimizeCAI(dna, ctable);
  }
  
  int *dna_seq = &dna[0];
  int length = dna.Length();
  int num_C = length / 3;

  ACMatcher<int, 4> f_matcher;
  ACMatcher<int, 4> d_matcher;
  int k;	// the maximum number of consecutive codons
       		// containing a motif

  int tmp = (int)std::ceil(BuildACMatcher(f_matcher, forbidden) / 3.0);
  k = tmp;
  tmp = (int)std::ceil(BuildACMatcher(d_matcher, desired) / 3.0);
  k = std::max(tmp,k);


  Counter codon_cntr(OD_MAX_CODONS_PER_AA, k);
   
  double **P = new double*[num_C];
  int **F = new int*[num_C];
  int **D = new int*[num_C];
  int ***Q = new int**[num_C];

   
  // now let's allocate a large hash for each DP matrix position
  // we do this instead of a multidimensional array
  unsigned int num_codon_combos = codon_cntr.IntMax() + 1;
   
  for (int i=0; i<num_C; i++) {
    P[i] = new double[num_codon_combos];
    F[i] = new int[num_codon_combos];
    D[i] = new int[num_codon_combos];
    Q[i] = new int*[num_codon_combos];
     
    for (int j=0; j<num_codon_combos; j++) {
      Q[i][j] = new int[OD_MAX_CODONS_PER_AA + 1];
    }
  }

  // Modified by GENETYX 2021.08.03
  int *cur_seq = new int[(k + 1) * OD_CODON_SIZE];
   
  // fill base case
  codon_cntr.Reset();
  while(codon_cntr.IntValue() < num_codon_combos) {
    double fitness = 1.0;
    bool invalid_codon_assignment = false;
    int hash_key = codon_cntr.IntValue();
     
    for(int i=0; i<k; i++) {
      int codon_index = codon_cntr.Digit(i);
      int base_pos = OD_CODON_SIZE * i;
      od_aa aa = ctable.GetAA4Codon(dna_seq[base_pos], dna_seq[base_pos + 1], 
                                    dna_seq[base_pos + 2]);

      // if this codon number exists for this aa
      if (codon_index < ctable.NumCodons4AA(aa)) {
        Codon codon = ctable.GetCodon4AA(aa, codon_index);
        CopyCodon( &cur_seq[base_pos], codon );
        fitness = fitness * codon.fit;	  
      } else {
        invalid_codon_assignment = true;
        break;
      }
    }
    
    if (invalid_codon_assignment) {
      // set this as an invalid entry in all DP matrices
      P[k - 1][hash_key] = OD_NINF;
      F[k - 1][hash_key] = OD_I_INF;
      D[k - 1][hash_key] = OD_I_NINF;
      Q[k - 1][hash_key][0] = 0;
    } else {
      // fill the DP matrices with motif matching statistics and fitness
      F[k - 1][hash_key] = f_matcher.Search(cur_seq, k * OD_CODON_SIZE);
      D[k - 1][hash_key] = d_matcher.Search(cur_seq, k * OD_CODON_SIZE);
      P[k - 1][hash_key] = fitness;
      // set the traceback count to 0 so we know to stop traceback later
      Q[k - 1][hash_key][0] = 0;
    }

    codon_cntr.Increment();	// try the next codon combination
  }
   
  // recursive case
  for (int i = k; i < num_C; i++) {

    codon_cntr.Reset();
    while(codon_cntr.IntValue() < num_codon_combos) {
      double fitness = 1.0;
      bool invalid_codon_assignment = false;
      int hash_key = codon_cntr.IntValue();
      Codon codon_i;

      P[i][hash_key] = OD_NINF;
      F[i][hash_key] = OD_I_INF;
      D[i][hash_key] = OD_I_NINF;
      Q[i][hash_key][0] = 0;

      // copy all but the earliest codon into cur_seq
       
      for(int j=0; j < k; j++) {
        int codon_index = codon_cntr.Digit(k-j-1);
        int base_pos = OD_CODON_SIZE * (i - j);
        od_aa aa = ctable.GetAA4Codon(dna_seq[base_pos], dna_seq[base_pos + 1], 
                                      dna_seq[base_pos + 2]);
         
        // if this codon number exists for this aa
        if (codon_index < ctable.NumCodons4AA(aa)) {
          Codon codon = ctable.GetCodon4AA(aa, codon_index);
          CopyCodon( &cur_seq[OD_CODON_SIZE * (k-j)], codon );
          fitness = fitness * codon.fit;

          if (j==0) {
            codon_i = codon;
          }
        } else {
          invalid_codon_assignment = true;
          break;
        }
      }

      if (!invalid_codon_assignment) {
        // now we try all possible codon assignments to position i - k
        int base_pos = OD_CODON_SIZE * (i - k);
        od_aa aa = ctable.GetAA4Codon(dna_seq[base_pos], dna_seq[base_pos + 1], 
                                      dna_seq[base_pos + 2]);
        int num_codons = ctable.NumCodons4AA(aa);
         
        Counter prev_codon_cntr(codon_cntr);
        prev_codon_cntr.LeftShift();
        int prev_hash_base = prev_codon_cntr.IntValue();
         
        for(int prev_index = 0; prev_index < num_codons; prev_index++) {
          int prev_hash_key = prev_hash_base + prev_index;
           
          if (P[i-1][prev_hash_key] == OD_NINF) {
            continue;
          } else {
            Codon codon = ctable.GetCodon4AA(aa, prev_index);
            CopyCodon(&cur_seq[0], codon);
             
            // determine the number of forbidden motifs ending in the last codon
            int new_forbidden = f_matcher.NumEndingInLast3(cur_seq, (k+1) * OD_CODON_SIZE);
            int num_forbidden = F[i-1][prev_hash_key] + new_forbidden;
             
            // determine the number of desired motifs ending in the last codon
            int new_desired = d_matcher.NumEndingInLast3(cur_seq, (k+1) * OD_CODON_SIZE);
            int num_desired = D[i-1][prev_hash_key] + new_desired;
             
            // First, we always pick selections with the least forbidden motifs
            if (num_forbidden < F[i][hash_key]) {
              P[i][hash_key] = codon_i.fit * P[i-1][prev_hash_key];
              Q[i][hash_key][0] = 1;
              Q[i][hash_key][1] = prev_index;
              F[i][hash_key] = num_forbidden;
              D[i][hash_key] = num_desired;
            }
            // Second, if the number of forbidden motifs is the same,
            // then pick the selection with the most desired motifs
            else if (num_forbidden == F[i][hash_key]) {
              if (num_desired > D[i][hash_key]) {
                P[i][hash_key] = codon_i.fit * P[i-1][prev_hash_key];
                Q[i][hash_key][0] = 1;
                Q[i][hash_key][1] = prev_index;
                D[i][hash_key] = num_desired;
              }
              // Third, if the number of forbidden and desired motifs are the same,
              // then pick the one with the highest fitness score
              else if (num_desired == D[i][hash_key]) {
                double local_score = codon_i.fit * P[i-1][prev_hash_key];
                // add all valid sequences here
                Q[i][hash_key][0] = Q[i][hash_key][0] + 1;
                Q[i][hash_key][ Q[i][hash_key][0] ] = prev_index;

                if (local_score > P[i][hash_key]) {
                  // always ensure the best score is stored in the first entry of the traceback array
                  P[i][hash_key] = codon_i.fit * P[i-1][prev_hash_key];
                  int tmp = Q[i][hash_key][1];
                  Q[i][hash_key][1] = prev_index;
                  Q[i][hash_key][ Q[i][hash_key][0] ] = tmp;
                }
              }
            }
          }
        }
      }
       
      codon_cntr.Increment();	// try the next codon combination
    }
  }

  // find optimal design based on the following criteria, in order:
  // 1) least number of forbidden motifs,
  // 2) most number of desired motifs,
  // 3) highest fitness score
  int min_forbidden = OD_I_INF;
  int max_desired = OD_I_NINF;
  double max_score = OD_NINF;
  Counter max_hash_cntr(codon_cntr);
  double score;
  int num_valid = 0;

  codon_cntr.Reset();
  while(codon_cntr.IntValue() < num_codon_combos) {
    int hash_key = codon_cntr.IntValue();
     
    // 1) Least number of forbidden motifs
    if (F[num_C-1][hash_key] < min_forbidden) {
      min_forbidden = F[num_C-1][hash_key];
      max_desired = D[num_C-1][hash_key];
      max_score = P[num_C-1][hash_key];
      max_hash_cntr = codon_cntr;
      num_valid = 1;

    } else if (F[num_C-1][hash_key] == min_forbidden) {
      // 2) Max number of desired motifs
      if (D[num_C-1][hash_key] > max_desired) {
        max_desired = D[num_C-1][hash_key];
        max_score = P[num_C-1][hash_key];
        max_hash_cntr = codon_cntr;
        num_valid = 1;
         
      } else if (D[num_C-1][hash_key] == max_desired) {
        num_valid++;
        // 3) Highest fitness score
        if (P[num_C-1][hash_key] > max_score)  {
          max_score = P[num_C-1][hash_key];
          max_hash_cntr = codon_cntr;
        }
      }
    }
    codon_cntr.Increment();
  }

  for(int j=0; j < k; j++) {
    int codon_index = max_hash_cntr.Digit(k-j-1);
    int base_pos = OD_CODON_SIZE * (num_C - 1 - j);
    od_aa aa = ctable.GetAA4Codon(dna_seq[base_pos], dna_seq[base_pos + 1], 
                                  dna_seq[base_pos + 2]);
         
    Codon codon = ctable.GetCodon4AA(aa, codon_index);
    CopyCodon( &dna_seq[base_pos], codon );
  }
     
  int i = num_C - 1;
  Counter tb_cntr(max_hash_cntr);
  int hash_key = max_hash_cntr.IntValue();
     
  while (Q[i][hash_key][0] != 0) {
    int q = 1;
       
    int codon_index = Q[i][hash_key][q];
    int base_pos = OD_CODON_SIZE * (i - k);
    od_aa aa = ctable.GetAA4Codon(dna_seq[base_pos], dna_seq[base_pos + 1], 
                                  dna_seq[base_pos + 2]);
         
    Codon codon = ctable.GetCodon4AA(aa, codon_index);
    CopyCodon( &dna_seq[base_pos], codon );
    tb_cntr.LeftShift();
    tb_cntr.SetDigit(0, codon_index);
    hash_key = tb_cntr.IntValue();
    i--;
  }
     
  max_score = std::exp(std::log(max_score) / (double)num_C); 
     
  // cleanup
  for (int i=0; i<num_C; i++) {
    for (int j=0; j<num_codon_combos; j++) {
      delete [] Q[i][j];
    }
    delete [] Q[i];
    delete [] D[i];
    delete [] F[i];
    delete [] P[i];
           
  }

  delete [] Q;
  delete [] D;
  delete [] F;
  delete [] P;
  
  // Modified by GENETYX 2021.08.03
  delete [] cur_seq;

  return max_score;
}

} // end of namespace oligodesign

#endif /* _CAI_OPTIMIZER_H_ */
