// This file implements a Codon Table representation.
// 
// Codon Optimizer - a reference implementation of the CAI optimization 
// and motif engineering algorithm by Anne Condon and Chris Thachuk. 
//
// Copyright (C) 2012 Chris Thachuk
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#ifndef _CODON_TABLE_H_
#define _CODON_TABLE_H_

#include <math.h>
#include <algorithm>
#include <string>
#include <utility>
#include <vector>

#include "codon.h"
#include "const.h"

namespace oligodesign {

namespace {

struct CodonCompare {
  bool operator () (Codon * p1, Codon * p2) const
  {
    return *p1 < *p2;
  }
};

inline void CopyCodon(int A[],Codon &B) {
  A[0] = B.b1; A[1] = B.b2; A[2] = B.b3;
}

} // end anonymous namespace


class CodonTable {
public:
  CodonTable() {
    // Modified by GENETYX 2021.08.04
    //SetupDefaultCodonBias();
  }

  double GetFreq(int b1, int b2, int b3) const {
    return codons[b1][b2][b3].freq;
  }

  double GetFit(int b1, int b2, int b3) const {
    return codons[b1][b2][b3].fit;
  }

  unsigned int NumCodons4AA(od_aa aa) const {
    return codon_freq[aa].size();
  }

  od_aa GetAA4Codon(int b1, int b2, int b3) const {
    return aa4codon[b1][b2][b3];
  }

  // TODO(cthachuk): add method to get best codon when passing in an od_aa parameter
  void SwapBest(int codon[]) const {
    int aa = aa4codon[codon[0]][codon[1]][codon[2]];
    Codon &c = *codon_freq[aa].back();
    CopyCodon(codon, c);
  }

  Codon GetCodon4AA(od_aa aa, int codon_index) const {
    return *(codon_freq[aa][codon_index]);
  }

  std::vector< Codon * > GetCodons4AA(od_aa aa) const {
    std::vector< Codon * > ret_codons;
    std::vector< Codon * >::const_iterator itr;
      
    for(itr = codon_freq[aa].begin(); itr != codon_freq[aa].end(); itr++) {
      ret_codons.push_back( *itr );
    }
      
    return ret_codons;
  }

  // Modified by GENETYX 2021.08.02
  bool ReadFromFile(const std::string& file);
  void SetupDefaultCodonBias();

  private:

    od_aa aa4codon[4][4][4];
    Codon codons[4][4][4];

    std::vector< Codon * > codon_freq[21]; // 20 plus stop codon
};

// Modified by GENETYX 2021.08.02
bool CodonTable::ReadFromFile(const std::string& file) {
  std::ifstream in(file);
  if (!in.is_open()) {
    return false;
  }

  od_aa aa4codon[4][4][4];

  auto cnvN = [](char c) {
    switch (c) {
    case 'T': return DNA_T;
    case 'A': return DNA_A;
    case 'C': return DNA_C;
    case 'G': return DNA_G;
    }
    assert(false);
    return DNA_C;
  };

  auto cnvA = [](const std::string& aa) {
    if (aa == "Ala") {
      return AA_ALA;
    }
    if (aa == "Arg") {
      return AA_ARG;
    }
    if (aa == "Asn") {
      return AA_ASN;
    }
    if (aa == "Asp") {
      return AA_ASP;
    }
    if (aa == "Cys") {
      return AA_CYS;
    }
    if (aa == "Gln") {
      return AA_GLN;
    }
    if (aa == "Glu") {
      return AA_GLU;
    }
    if (aa == "Gly") {
      return AA_GLY;
    }
    if (aa == "His") {
      return AA_HIS;
    }
    if (aa == "Ile") {
      return AA_ILE;
    }
    if (aa == "Leu") {
      return AA_LEU;
    }
    if (aa == "Lys") {
      return AA_LYS;
    }
    if (aa == "Met") {
      return AA_MET;
    }
    if (aa == "Phe") {
      return AA_PHE;
    }
    if (aa == "Pro") {
      return AA_PRO;
    }
    if (aa == "Ser") {
      return AA_SER;
    }
    if (aa == "Thr") {
      return AA_THR;
    }
    if (aa == "Trp") {
      return AA_TRP;
    }
    if (aa == "Tyr") {
      return AA_TYR;
    }
    if (aa == "Val") {
      return AA_VAL;
    }
    if (aa == "End") {
      return AA_END;
    }
    assert(false);
    return AA_END;
  };

  struct CodonInfo {
    std::string codon;
    std::string aa;
    int count;
  };

  std::map<std::string, std::vector<CodonInfo>> m;

  std::string line;
  while (std::getline(in, line)) {
    std::vector<std::string> v;
    boost::algorithm::split(v, line, boost::is_any_of(","));
    if (v.size() != 3 && (v[0].size() != 3 || v[1].size() != 3 || v[2].size() != 3)) {
      return false;
    }
    const auto codon = v[0];
    const auto aa = v[1];
    const auto count = std::stoi(v[2]);
    aa4codon[cnvN(codon[0])][cnvN(codon[1])][cnvN(codon[2])] = cnvA(aa);
    m[aa].push_back(CodonInfo{codon,aa,count});
  }

  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 4; j++)
      for (int k = 0; k < 4; k++)
        this->aa4codon[i][j][k] = aa4codon[i][j][k];

  for (const auto& p : m) {
    int sum = 0;
    for (const auto& c : p.second) {
      sum += c.count;
    }
    double maxFreq = 0.0;
    std::vector<Codon> cs;
    for (const auto& c : p.second) {
      auto freq = static_cast<double>(c.count) / static_cast<double>(sum);
      maxFreq = std::max(freq, maxFreq);
      cs.push_back(Codon(c.codon.c_str(), freq, 0.0));
    }
    for (auto& c : cs) {
      c.fit = c.freq / maxFreq;
    }
    for (auto& c : cs) {
      codons[c.b1][c.b2][c.b3] = c;
    }
  }

  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < 4; j++) {
      for (int k = 0; k < 4; k++) {
        codon_freq[aa4codon[i][j][k]].push_back(&codons[i][j][k]);
      }
    }
  }

  for (int i = 0; i < 21; i++) {
    double total_freq = 0.0;
    for (std::vector<Codon*>::iterator itr = codon_freq[i].begin();
         itr != codon_freq[i].end(); itr++) {
      total_freq += (*itr)->freq;
    }
    for (std::vector<Codon*>::iterator itr = codon_freq[i].begin();
         itr != codon_freq[i].end(); itr++) {
      (*itr)->freq = (*itr)->freq / total_freq;
    }
    std::sort(codon_freq[i].begin(), codon_freq[i].end(), CodonCompare());
  }

  return true;
}

void CodonTable::SetupDefaultCodonBias() {	
  od_aa aa4codon[4][4][4] = {
    {/* First position 'A' */
      {/* Second position 'A' */
        /* Third position 'A' */  AA_LYS,
        /* Third position 'C' */  AA_ASN,
        /* Third position 'G' */  AA_LYS,
        /* Third position 'T' */  AA_ASN
      },
      { AA_THR, AA_THR, AA_THR, AA_THR },
      { AA_ARG, AA_SER, AA_ARG, AA_SER },
      { AA_ILE, AA_ILE, AA_MET, AA_ILE }
    },
    {/* First position 'C' */
      { AA_GLN, AA_HIS, AA_GLN, AA_HIS },
      { AA_PRO, AA_PRO, AA_PRO, AA_PRO },
      { AA_ARG, AA_ARG, AA_ARG, AA_ARG },
      { AA_LEU, AA_LEU, AA_LEU, AA_LEU }
    },
    {/* First position 'G' */
      { AA_GLU, AA_ASP, AA_GLU, AA_ASP },
      { AA_ALA, AA_ALA, AA_ALA, AA_ALA },
      { AA_GLY, AA_GLY, AA_GLY, AA_GLY },
      { AA_VAL, AA_VAL, AA_VAL, AA_VAL }
    },
    {/* First position 'T' */
      { AA_END, AA_TYR, AA_END, AA_TYR },
      { AA_SER, AA_SER, AA_SER, AA_SER },
      { AA_END, AA_CYS, AA_TRP, AA_CYS },
      { AA_LEU, AA_PHE, AA_LEU, AA_PHE }
    }
  };

  for (int i = 0; i < 4; i++)
    for (int j = 0; j < 4; j++)
      for (int k = 0; k < 4; k++)
        this->aa4codon[i][j][k] = aa4codon[i][j][k];

  codons[DNA_G][DNA_G][DNA_G] = Codon("GGG", 0.16, 0.48);
  codons[DNA_G][DNA_G][DNA_A] = Codon("GGA", 0.16, 0.48);
  codons[DNA_G][DNA_G][DNA_T] = Codon("GGT", 0.34, 1.00);
  codons[DNA_G][DNA_G][DNA_C] = Codon("GGC", 0.34, 1.00);
  codons[DNA_G][DNA_A][DNA_G] = Codon("GAG", 0.34, 0.52);
  codons[DNA_G][DNA_A][DNA_A] = Codon("GAA", 0.66, 1.00);
  codons[DNA_G][DNA_A][DNA_T] = Codon("GAT", 0.63, 1.00);
  codons[DNA_G][DNA_A][DNA_C] = Codon("GAC", 0.37, 0.58);
  codons[DNA_G][DNA_T][DNA_G] = Codon("GTG", 0.32, 1.00);
  codons[DNA_G][DNA_T][DNA_A] = Codon("GTA", 0.18, 0.54);
  codons[DNA_G][DNA_T][DNA_T] = Codon("GTT", 0.30, 0.91);
  codons[DNA_G][DNA_T][DNA_C] = Codon("GTC", 0.20, 0.62);
  codons[DNA_G][DNA_C][DNA_G] = Codon("GCG", 0.30, 1.00);
  codons[DNA_G][DNA_C][DNA_A] = Codon("GCA", 0.24, 0.83);
  codons[DNA_G][DNA_C][DNA_T] = Codon("GCT", 0.20, 0.68);
  codons[DNA_G][DNA_C][DNA_C] = Codon("GCC", 0.26, 0.90);
  codons[DNA_A][DNA_G][DNA_G] = Codon("AGG", 0.05, 0.17);
  codons[DNA_A][DNA_G][DNA_A] = Codon("AGA", 0.09, 0.28);
  codons[DNA_A][DNA_G][DNA_T] = Codon("AGT", 0.17, 0.76);
  codons[DNA_A][DNA_G][DNA_C] = Codon("AGC", 0.22, 1.00);
  codons[DNA_A][DNA_A][DNA_G] = Codon("AAG", 0.28, 0.39);
  codons[DNA_A][DNA_A][DNA_A] = Codon("AAA", 0.72, 1.00);
  codons[DNA_A][DNA_A][DNA_T] = Codon("AAT", 0.53, 1.00);
  codons[DNA_A][DNA_A][DNA_C] = Codon("AAC", 0.47, 0.87);
  codons[DNA_A][DNA_T][DNA_G] = Codon("ATG", 1.00, 1.00);
  codons[DNA_A][DNA_T][DNA_A] = Codon("ATA", 0.15, 0.32);
  codons[DNA_A][DNA_T][DNA_T] = Codon("ATT", 0.48, 1.00);
  codons[DNA_A][DNA_T][DNA_C] = Codon("ATC", 0.36, 0.76);
  codons[DNA_A][DNA_C][DNA_G] = Codon("ACG", 0.24, 0.66);
  codons[DNA_A][DNA_C][DNA_A] = Codon("ACA", 0.21, 0.57);
  codons[DNA_A][DNA_C][DNA_T] = Codon("ACT", 0.20, 0.55);
  codons[DNA_A][DNA_C][DNA_C] = Codon("ACC", 0.36, 1.00);
  codons[DNA_T][DNA_G][DNA_G] = Codon("TGG", 1.00, 1.00);
  codons[DNA_T][DNA_G][DNA_A] = Codon("TGA", 0.32, 0.55);
  codons[DNA_T][DNA_G][DNA_T] = Codon("TGT", 0.48, 0.92);
  codons[DNA_T][DNA_G][DNA_C] = Codon("TGC", 0.52, 1.00);
  codons[DNA_T][DNA_A][DNA_G] = Codon("TAG", 0.09, 0.15);
  codons[DNA_T][DNA_A][DNA_A] = Codon("TAA", 0.59, 1.00);
  codons[DNA_T][DNA_A][DNA_T] = Codon("TAT", 0.61, 1.00);
  codons[DNA_T][DNA_A][DNA_C] = Codon("TAC", 0.39, 0.64);
  codons[DNA_T][DNA_T][DNA_G] = Codon("TTG", 0.13, 0.29);
  codons[DNA_T][DNA_T][DNA_A] = Codon("TTA", 0.15, 0.34);
  codons[DNA_T][DNA_T][DNA_T] = Codon("TTT", 0.60, 1.00);
  codons[DNA_T][DNA_T][DNA_C] = Codon("TTC", 0.40, 0.68);
  codons[DNA_T][DNA_C][DNA_G] = Codon("TCG", 0.13, 0.56);
  codons[DNA_T][DNA_C][DNA_A] = Codon("TCA", 0.16, 0.72);
  codons[DNA_T][DNA_C][DNA_T] = Codon("TCT", 0.17, 0.77);
  codons[DNA_T][DNA_C][DNA_C] = Codon("TCC", 0.14, 0.64);
  codons[DNA_C][DNA_G][DNA_G] = Codon("CGG", 0.13, 0.39);
  codons[DNA_C][DNA_G][DNA_A] = Codon("CGA", 0.08, 0.24);
  codons[DNA_C][DNA_G][DNA_T] = Codon("CGT", 0.33, 1.00);
  codons[DNA_C][DNA_G][DNA_C] = Codon("CGC", 0.32, 0.96);
  codons[DNA_C][DNA_A][DNA_G] = Codon("CAG", 0.67, 1.00);
  codons[DNA_C][DNA_A][DNA_A] = Codon("CAA", 0.33, 0.50);
  codons[DNA_C][DNA_A][DNA_T] = Codon("CAT", 0.59, 1.00);
  codons[DNA_C][DNA_A][DNA_C] = Codon("CAC", 0.41, 0.70);
  codons[DNA_C][DNA_T][DNA_G] = Codon("CTG", 0.44, 1.00);
  codons[DNA_C][DNA_T][DNA_A] = Codon("CTA", 0.05, 0.11);
  codons[DNA_C][DNA_T][DNA_T] = Codon("CTT", 0.13, 0.29);
  codons[DNA_C][DNA_T][DNA_C] = Codon("CTC", 0.10, 0.23);
  codons[DNA_C][DNA_C][DNA_G] = Codon("CCG", 0.45, 1.00);
  codons[DNA_C][DNA_C][DNA_A] = Codon("CCA", 0.21, 0.48);
  codons[DNA_C][DNA_C][DNA_T] = Codon("CCT", 0.20, 0.46);
  codons[DNA_C][DNA_C][DNA_C] = Codon("CCC", 0.14, 0.32);
      
  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < 4; j++) {
      for (int k = 0; k < 4; k++) {
        codon_freq[aa4codon[i][j][k]].push_back(&codons[i][j][k]);
      }
    }
  }

  for (int i = 0; i < 21; i++) {
    double total_freq = 0.0;
    for (std::vector< Codon * >::iterator itr = codon_freq[i].begin();
         itr != codon_freq[i].end(); itr++) {
      total_freq += (*itr)->freq;
    }
    for (std::vector< Codon * >::iterator itr = codon_freq[i].begin();
         itr != codon_freq[i].end(); itr++) {
      (*itr)->freq = (*itr)->freq / total_freq;
    }
    sort(codon_freq[i].begin(), codon_freq[i].end(), CodonCompare());
  }
}

} // end namespace oligodesign

#endif /* _CODON_TABLE_H_ */

