// This file implements a simple DNA sequence convenience class.
// 
// Codon Optimizer - a reference implementation of the CAI optimization 
// and motif engineering algorithm by Anne Condon and Chris Thachuk. 
//
// Copyright (C) 2012 Chris Thachuk
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.


#ifndef _DNA_SEQUENCE_H_
#define _DNA_SEQUENCE_H_

#include <ostream>
#include <string>
#include <vector>

#include <boost/algorithm/string.hpp>

#include "const.h"

namespace oligodesign {

class DNASequence {
public:
  DNASequence()
    :_seq(), _valid_seq(true)
  {}

  DNASequence(const std::string &seq)
    :_valid_seq(true) {
    SetSequence(seq);
  }

  size_t Length() const {
    return _seq.size();
  }

  bool ValidSeq() const {
    return _valid_seq;
  }

  const int& operator[] (size_t i) const {
    return _seq[i];
  }

  int& operator[] (size_t i) {
    return _seq[i];
  }

 private:
  std::vector<int> _seq;
  bool _valid_seq;

  void SetSequence(const std::string &seq) {
    std::string buf(seq);
    boost::to_upper(buf);
    _seq.clear();
    _seq.resize(buf.length());
    
    for (size_t i = 0; i < buf.length(); ++i) {
      switch (buf[i]) {
      case 'A':
        _seq[i] = DNA_A;
        break;
      case 'T':
        _seq[i] = DNA_T;
        break;
      case 'G':
        _seq[i] = DNA_G;
        break;
      case 'C':
        _seq[i] = DNA_C;
        break;
      default: // note: unknown characters will silently become 'C'
        _seq[i] = DNA_C;
        _valid_seq = false;
        break;
      }
    }
  }

};

} // end namespace oligodesign

std::ostream& operator<<(std::ostream& out, const oligodesign::DNASequence& dna) {
  using namespace oligodesign;

  for(size_t i = 0; i < dna.Length(); ++i) {
    char c = 'C';
    switch (dna[i]) {
    case DNA_A:
      c = 'A';
      break;
    case DNA_T:
      c = 'T';
      break;
    case DNA_G:
      c = 'G';
      break;
    case DNA_C:
      c = 'C';
      break;
    }
    out << c;
  }
  return out;
}


#endif /* _DNA_SEQUENCE_H_ */
